from di.executors._async import AsyncExecutor, ConcurrentAsyncExecutor
from di.executors._sync import SyncExecutor

__all__ = (
    "AsyncExecutor",
    "ConcurrentAsyncExecutor",
    "SyncExecutor",
)



import contextvars
from typing import Any, Awaitable, Callable, TypeVar

try:
    import anyio
except ImportError as e:
    raise ImportError(
        "Using AsyncExector or ConcurrentAsyncExecutor requires installing anyio"
        " (`pip install anyio`) or the anyio extra (`pip install di[anyio]`)"
    ) from e
import anyio.abc

from di.api.executor import StateType, SupportsAsyncExecutor, SupportsTaskGraph, Task

T = TypeVar("T")


def callable_in_thread_pool(call: Callable[..., T]) -> Callable[..., Awaitable[T]]:
    def inner(*args: Any, **kwargs: Any) -> "Awaitable[T]":
        return anyio.to_thread.run_sync(
            contextvars.copy_context().run, lambda: call(*args, **kwargs)
        )  # type: ignore[return-value]

    return inner


async def _execute_task(task: Task[StateType], state: StateType) -> None:
    if getattr(task.dependant, "sync_to_thread", False):
        await callable_in_thread_pool(task.compute)(state)
    else:
        maybe_aw = task.compute(state)
        if maybe_aw is not None:
            await maybe_aw


class AsyncExecutor(SupportsAsyncExecutor):
    """An executor that executes sync and async dependencies sequentially."""

    async def execute_async(
        self, tasks: SupportsTaskGraph[StateType], state: StateType
    ) -> None:
        for task in tasks.static_order():
            await _execute_task(task, state)


async def async_worker(
    task: Task[StateType],
    tasks: SupportsTaskGraph[StateType],
    state: StateType,
    taskgroup: anyio.abc.TaskGroup,
) -> None:
    await _execute_task(task, state)
    tasks.done(task)
    for task in tasks.get_ready():
        taskgroup.start_soon(async_worker, task, tasks, state, taskgroup)


class ConcurrentAsyncExecutor(SupportsAsyncExecutor):
    async def execute_async(
        self, tasks: SupportsTaskGraph[StateType], state: StateType
    ) -> None:
        async with anyio.create_task_group() as taskgroup:
            for task in tasks.get_ready():
                taskgroup.start_soon(async_worker, task, tasks, state, taskgroup)



import typing
from typing import Any, Generic, Iterable, Mapping, TypeVar

from di.api.dependencies import DependantBase, DependencyParameter

Dependency = Any

DependencyType = TypeVar("DependencyType")


class SolvedDependant(Generic[DependencyType]):
    """Representation of a fully solved dependency as DAG.

    A SolvedDependant could be a user's endpoint/controller function.
    """

    __slots__ = ("dependency", "dag", "container_cache")

    dependency: DependantBase[DependencyType]
    dag: Mapping[DependantBase[Any], Iterable[DependencyParameter]]
    # container_cache can be used by the creating container to store data that is tied
    # to the SolvedDependant
    container_cache: typing.Any

    def __init__(
        self,
        dependency: DependantBase[DependencyType],
        dag: Mapping[DependantBase[Any], Iterable[DependencyParameter]],
        container_cache: typing.Any,
    ):
        self.dependency = dependency
        self.dag = dag
        self.container_cache = container_cache



from di.api.executor import StateType, SupportsSyncExecutor, SupportsTaskGraph


class SyncExecutor(SupportsSyncExecutor):
    """An executor that executes only sync dependencies.

    Dependencies are executed sequentially.
    If any async dependencies are encountered a RuntimeError will be raised.
    If there are no async dependencies, this will be faster than using `AsyncExecutor` because there is no event loop overhead.
    """

    def execute_sync(
        self, tasks: SupportsTaskGraph[StateType], state: StateType
    ) -> None:
        for task in tasks.static_order():
            maybe_aw = task.compute(state)
            if maybe_aw is not None:
                raise TypeError("Cannot execute async dependencies in execute_sync")



import inspect
import sys
from typing import Any, Optional

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import Protocol
else:  # pragma: no cover
    from typing import Protocol


from di.api.dependencies import DependantBase


class BindHook(Protocol):
    def __call__(
        self, param: Optional[inspect.Parameter], dependant: DependantBase[Any]
    ) -> Optional[DependantBase[Any]]:  # pragma: no cover
        ...



from __future__ import annotations

import contextlib
from contextlib import AsyncExitStack, ExitStack
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    TypeVar,
    Union,
)

from di._utils.inspect import (
    is_async_gen_callable,
    is_coroutine_callable,
    is_gen_callable,
)
from di._utils.scope_map import ScopeMap
from di._utils.types import CacheKey
from di.api.dependencies import DependantBase
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.exceptions import IncompatibleDependencyError


class ExecutionState(NamedTuple):
    stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]]
    results: List[Any]
    cache: ScopeMap[CacheKey, Any]
    values: Mapping[DependencyProvider, Any]


DependencyType = TypeVar("DependencyType")


UNSET: Any = object()


class Task:
    __slots__ = (
        "wrapped_call",
        "user_function",
        "scope",
        "cache_key",
        "dependant",
        "task_id",
        "call_user_func_with_deps",
        "compute",
    )

    compute: Union[
        Callable[[ExecutionState], Awaitable[None]],
        Callable[[ExecutionState], None],
    ]
    wrapped_call: DependencyProvider
    user_function: DependencyProvider

    def __init__(
        self,
        scope: Scope,
        call: DependencyProvider,
        use_cache: bool,
        cache_key: CacheKey,
        dependant: DependantBase[Any],
        task_id: int,
        positional_parameters: Iterable[Task],
        keyword_parameters: Mapping[str, Task],
    ) -> None:
        self.scope = scope
        self.user_function = call
        self.cache_key = cache_key
        self.dependant = dependant
        self.task_id = task_id
        if is_async_gen_callable(self.user_function):
            self.wrapped_call = contextlib.asynccontextmanager(call)  # type: ignore[arg-type]
            if use_cache:
                self.compute = self.compute_async_cm_cache
            else:
                self.compute = self.compute_async_cm_no_cache
        elif is_coroutine_callable(self.user_function):
            self.wrapped_call = self.user_function
            if use_cache:
                self.compute = self.compute_async_coro_cache
            else:
                self.compute = self.compute_async_coro_no_cache
        elif is_gen_callable(call):
            self.wrapped_call = contextlib.contextmanager(call)  # type: ignore[arg-type]
            if use_cache:
                self.compute = self.compute_sync_cm_cache
            else:
                self.compute = self.compute_sync_cm_no_cache
        else:
            self.wrapped_call = call
            if use_cache:
                self.compute = self.compute_sync_func_cache
            else:
                self.compute = self.compute_sync_func_no_cache

        self.call_user_func_with_deps = self.generate_execute_fn(
            self.wrapped_call, positional_parameters, keyword_parameters
        )

    def __hash__(self) -> int:
        return self.task_id

    @staticmethod
    def generate_execute_fn(
        call: DependencyProvider,
        positional_parameters: Iterable[Task],
        keyword_parameters: Mapping[str, Task],
    ) -> Callable[[List[Any]], Any]:
        # this codegen speeds up argument collection and passing
        # by avoiding creation of intermediary containers to store the values
        positional_arg_template = "results[{}]"
        keyword_arg_template = "{}=results[{}]"
        args: List[str] = []
        for task in positional_parameters:
            args.append(positional_arg_template.format(task.task_id))
        for keyword, task in keyword_parameters.items():
            args.append(keyword_arg_template.format(keyword, task.task_id))
        locals: Dict[str, Any] = {}
        globals = {"call": call}
        exec(f'def execute(results): return call({",".join(args)})', globals, locals)
        return locals["execute"]  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(scope={self.scope}, call={self.user_function})"
        )

    async def compute_async_coro_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        value = state.cache.get_key(self.cache_key, scope=self.scope, default=UNSET)
        if value is not UNSET:
            state.results[self.task_id] = value
            return
        dependency_value = await self.call_user_func_with_deps(state.results)
        state.results[self.task_id] = dependency_value
        state.cache.set(self.cache_key, dependency_value, scope=self.scope)

    async def compute_async_coro_no_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        dependency_value = await self.call_user_func_with_deps(state.results)
        state.results[self.task_id] = dependency_value

    async def compute_async_cm_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        value = state.cache.get_key(self.cache_key, scope=self.scope, default=UNSET)
        if value is not UNSET:
            state.results[self.task_id] = value
            return
        try:
            enter = state.stacks[self.scope].enter_async_context  # type: ignore[union-attr]
        except AttributeError:
            raise IncompatibleDependencyError(
                f"The dependency {self.user_function} is an awaitable dependency"
                f" and cannot be used in the sync scope {self.scope}"
            ) from None

        dependency_value: Any = await enter(
            self.call_user_func_with_deps(state.results)
        )
        state.results[self.task_id] = dependency_value
        state.cache.set(self.cache_key, dependency_value, scope=self.scope)

    async def compute_async_cm_no_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        try:
            enter = state.stacks[self.scope].enter_async_context  # type: ignore[union-attr]
        except AttributeError:
            raise IncompatibleDependencyError(
                f"The dependency {self.user_function} is an awaitable dependency"
                f" and cannot be used in the sync scope {self.scope}"
            ) from None
        dependency_value: Any = await enter(
            self.call_user_func_with_deps(state.results)
        )
        state.results[self.task_id] = dependency_value

    def compute_sync_cm_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        value = state.cache.get_key(self.cache_key, scope=self.scope, default=UNSET)
        if value is not UNSET:
            state.results[self.task_id] = value
            return
        val = state.stacks[self.scope].enter_context(
            self.call_user_func_with_deps(state.results)
        )
        state.results[self.task_id] = val
        state.cache.set(self.cache_key, val, scope=self.scope)

    def compute_sync_cm_no_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        state.results[self.task_id] = state.stacks[self.scope].enter_context(
            self.call_user_func_with_deps(state.results)
        )

    def compute_sync_func_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return
        value = state.cache.get_key(self.cache_key, scope=self.scope, default=UNSET)
        if value is not UNSET:
            state.results[self.task_id] = value
            return
        val = self.call_user_func_with_deps(state.results)
        state.results[self.task_id] = val
        state.cache.set(self.cache_key, val, scope=self.scope)

    def compute_sync_func_no_cache(self, state: ExecutionState) -> None:
        if self.user_function in state.values:
            state.results[self.task_id] = state.values[self.user_function]
            return

        state.results[self.task_id] = self.call_user_func_with_deps(state.results)



import inspect
import sys
from typing import Any, Generic, List, NamedTuple, Optional, TypeVar

if sys.version_info < (3, 8):
    from typing_extensions import Protocol, runtime_checkable
else:
    from typing import Protocol, runtime_checkable

from di._utils.types import CacheKey
from di.api.providers import DependencyProviderType
from di.api.scopes import Scope

T = TypeVar("T")


__all__ = (
    "CacheKey",
    "DependantBase",
    "DependencyParameter",
    "InjectableClassProvider",
)


@runtime_checkable
class InjectableClassProvider(Protocol):
    @classmethod
    def __di_dependency__(cls, param: inspect.Parameter) -> "DependantBase[Any]":
        ...


class DependantBase(Generic[T]):
    """A dependant is an object that can provide the container with:
    - A hash, to compare itself against other dependants
    - A scope
    - A callable who's returned value is the dependency
    """

    call: Optional[DependencyProviderType[T]]
    scope: Scope
    use_cache: bool

    @property
    def cache_key(self) -> CacheKey:
        raise NotImplementedError

    def get_dependencies(self) -> "List[DependencyParameter]":
        """Collect all of the sub dependencies for this dependant"""
        raise NotImplementedError


class DependencyParameter(NamedTuple):
    dependency: DependantBase[Any]
    parameter: Optional[inspect.Parameter]



import inspect
from typing import Any, Optional

from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.dependant._dependant import Dependant


class Injectable:
    __slots__ = ()

    def __init_subclass__(
        cls,
        call: Optional[DependencyProvider] = None,
        scope: Scope = None,
        use_cache: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)

        def create_dependant(cls_: Any, param: inspect.Parameter) -> Dependant[Any]:
            return Dependant(call or cls_, scope=scope, use_cache=use_cache)

        cls.__di_dependency__ = classmethod(create_dependant)  # type: ignore



from __future__ import annotations

from typing import Any, Iterable, List, TypeVar

from di.api.dependencies import CacheKey, DependantBase, DependencyParameter

T = TypeVar("T")


class JoinedDependant(DependantBase[T]):
    """A Dependant that aggregates other dependants without directly depending on them"""

    __slots__ = ("dependant", "siblings")

    def __init__(
        self,
        dependant: DependantBase[T],
        *,
        siblings: Iterable[DependantBase[Any]],
    ) -> None:
        self.call = dependant.call
        self.dependant = dependant
        self.siblings = siblings
        self.scope = dependant.scope
        self.use_cache = dependant.use_cache

    def get_dependencies(self) -> List[DependencyParameter]:
        """Get the dependencies of our main dependant and all siblings"""
        return [
            *self.dependant.get_dependencies(),
            *(DependencyParameter(dep, None) for dep in self.siblings),
        ]

    @property
    def cache_key(self) -> CacheKey:
        return (self.dependant.cache_key, tuple((s.cache_key for s in self.siblings)))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dependant={self.dependant}, siblings={self.siblings})"



from __future__ import annotations

import sys
from typing import Awaitable, Callable, Hashable, Iterable, TypeVar, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from di.api.dependencies import DependantBase

StateType = TypeVar("StateType")


class Task(Hashable, Protocol[StateType]):
    dependant: DependantBase[StateType]

    @property
    def compute(self) -> Callable[[StateType], Union[Awaitable[None], None]]:
        ...


class SupportsTaskGraph(Protocol[StateType]):
    def done(self, task: Task[StateType]) -> None:
        ...

    def get_ready(self) -> Iterable[Task[StateType]]:
        ...

    def is_active(self) -> bool:
        ...

    def static_order(self) -> Iterable[Task[StateType]]:
        ...


class SupportsSyncExecutor(Protocol):
    def execute_sync(
        self, tasks: SupportsTaskGraph[StateType], state: StateType
    ) -> None:
        raise NotImplementedError


class SupportsAsyncExecutor(Protocol):
    async def execute_async(
        self, tasks: SupportsTaskGraph[StateType], state: StateType
    ) -> None:
        raise NotImplementedError



from contextlib import AsyncExitStack, ExitStack
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from di._utils.scope_map import ScopeMap
from di._utils.types import CacheKey, FusedContextManager
from di.api.scopes import Scope


class ContainerState:
    __slots__ = ("cached_values", "stacks")

    def __init__(
        self,
        cached_values: Optional[ScopeMap[CacheKey, Any]] = None,
        stacks: Optional[Dict[Scope, Union[AsyncExitStack, ExitStack]]] = None,
    ) -> None:
        self.cached_values = cached_values or ScopeMap()
        self.stacks = stacks or {}

    def enter_scope(self, scope: Scope) -> "FusedContextManager[ContainerState]":
        """Enter a scope and get back a new ContainerState object that you can use to execute dependencies."""
        new = ContainerState(
            cached_values=ScopeMap(self.cached_values.copy()),
            stacks=self.stacks.copy(),
        )
        return ScopeContext(new, scope)


class ScopeContext(FusedContextManager[ContainerState]):
    __slots__ = ("state", "scope", "stack")
    stack: Union[AsyncExitStack, ExitStack]

    def __init__(self, state: ContainerState, scope: Scope) -> None:
        self.state = state
        self.scope = scope

    def __enter__(self) -> ContainerState:
        self.state.stacks[self.scope] = self.stack = ExitStack()
        self.state.cached_values.add_scope(self.scope)
        return self.state

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Union[None, bool]:
        return self.stack.__exit__(exc_type, exc_value, traceback)  # type: ignore[union-attr,no-any-return]

    async def __aenter__(self) -> ContainerState:
        self.state.stacks[self.scope] = self.stack = AsyncExitStack()
        self.state.cached_values.add_scope(self.scope)
        return self.state

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Union[None, bool]:
        return await self.stack.__aexit__(exc_type, exc_value, traceback)  # type: ignore[union-attr,no-any-return]



from contextlib import contextmanager
from typing import (
    Any,
    ContextManager,
    Generator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
)

from di._utils.types import FusedContextManager
from di.api.dependencies import DependantBase
from di.api.executor import SupportsAsyncExecutor, SupportsSyncExecutor
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.api.solved import SolvedDependant
from di.container._bind_hook import BindHook
from di.container._execution_planning import plan_execution
from di.container._solving import ScopeResolver, solve
from di.container._state import ContainerState

DependencyType = TypeVar("DependencyType")


class Container:
    """Solve and execute dependencies.

    Generally you will want one Container per application.
    There is not performance advantage to re-using a container, the only reason to do so is to share binds.
    For each "thing" you want to wire with di and execute you'll want to call `Container.solve()`
    exactly once and then keep a reference to the returned `SolvedDependant` to pass to `Container.execute`.
    Solving is very expensive so avoid doing it in a hot loop.
    """

    __slots__ = ("_bind_hooks", "_state")

    _bind_hooks: List[BindHook]

    def __init__(self) -> None:
        self._bind_hooks = []

    def bind(
        self,
        hook: BindHook,
    ) -> ContextManager[None]:
        """Replace a dependency provider with a new one.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).
        """

        self._bind_hooks.append(hook)

        @contextmanager
        def unbind() -> "Generator[None, None, None]":
            try:
                yield
            finally:
                self._bind_hooks.remove(hook)

        return unbind()

    def solve(
        self,
        dependency: DependantBase[DependencyType],
        scopes: Sequence[Scope],
        scope_resolver: Optional[ScopeResolver] = None,
    ) -> SolvedDependant[DependencyType]:
        """Build the dependency graph.

        Should happen once, maybe during startup.

        Solving dependencies can be slow.
        """
        return solve(dependency, scopes, self._bind_hooks, scope_resolver)

    def execute_sync(
        self,
        solved: SolvedDependant[DependencyType],
        executor: SupportsSyncExecutor,
        *,
        state: ContainerState,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        This method is synchronous and uses a synchronous executor,
        but the executor may still be able to execute async dependencies.
        """
        results, ts, execution_state, root_task = plan_execution(
            stacks=state.stacks,
            cache=state.cached_values,
            solved=solved,
            values=values,
        )
        executor.execute_sync(ts, execution_state)
        return results[root_task.task_id]  # type: ignore[no-any-return]

    async def execute_async(
        self,
        solved: SolvedDependant[DependencyType],
        executor: SupportsAsyncExecutor,
        *,
        state: ContainerState,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency."""
        results, ts, execution_state, root_task = plan_execution(
            stacks=state.stacks,
            cache=state.cached_values,
            solved=solved,
            values=values,
        )
        await executor.execute_async(ts, execution_state)
        return results[root_task.task_id]  # type: ignore[no-any-return]

    def enter_scope(
        self, scope: Scope, state: Optional[ContainerState] = None
    ) -> FusedContextManager[ContainerState]:
        state = state or ContainerState()
        return state.enter_scope(scope)



from di.dependant._dependant import Dependant, Marker
from di.dependant._injectable_type import Injectable
from di.dependant._joined import JoinedDependant

__all__ = (
    "Marker",
    "Dependant",
    "JoinedDependant",
    "Injectable",
)



from di.container._bind_by_type_hook import bind_by_type
from di.container._bind_hook import BindHook
from di.container._container import Container
from di.container._state import ContainerState

__all__ = ("BindHook", "Container", "ContainerState", "bind_by_type")



import inspect
import sys
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, TypeVar

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from graphlib2 import TopologicalSorter

from di._utils.task import Task
from di.api.dependencies import CacheKey, DependantBase, DependencyParameter
from di.api.scopes import Scope
from di.api.solved import SolvedDependant
from di.container._bind_hook import BindHook
from di.container._execution_planning import SolvedDependantCache
from di.exceptions import (
    DependencyCycleError,
    ScopeViolationError,
    SolvingError,
    UnknownScopeError,
    WiringError,
)

T = TypeVar("T")


POSITIONAL_PARAMS = (
    inspect.Parameter.POSITIONAL_ONLY,
    inspect.Parameter.POSITIONAL_OR_KEYWORD,
)


class ScopeResolver(Protocol):
    def __call__(
        self,
        __dependant: DependantBase[Any],
        __sub_dependenant_scopes: Sequence[Scope],
        __solver_scopes: Sequence[Scope],
    ) -> Scope:
        """Infer scopes for a Marker/Dependant that does not have an explicit scope.

        The three paramters given are:
        - `sub_dependenant_scopes`: the scopes of all sub-dependencies (if any).
          This can be used to set a lower bound for the scope.
          For example, if a sub dependency has some "singleton" scope
          our current dependency (the `dependant` argument) cannot have some "ephemeral"
          scope because that would violate scoping rules.
        - `solver_scopes`: the scopes passed to `Container.solve`. Provided for convenience.
        - `dependant`: the current dependency we are inferring a scope for.
        """


def get_path_str(path: Iterable[DependantBase[Any]]) -> str:
    return " -> ".join(
        [repr(item) if item.call is not None else repr(item.call) for item in path]
    )


def get_params(
    dep: DependantBase[Any],
    binds: Iterable[BindHook],
    path: Iterable[DependantBase[Any]],
) -> List[DependencyParameter]:
    """Get Dependants for parameters and resolve binds"""
    params = dep.get_dependencies().copy()
    for idx, param in enumerate(params):
        for hook in binds:
            match = hook(param.parameter, param.dependency)
            if match is not None:
                param = param._replace(dependency=match)
        params[idx] = param
        if param.parameter is not None:
            if (
                param.dependency.call is None
                and param.parameter.default is param.parameter.empty
            ):
                raise WiringError(
                    (
                        f"The parameter {param.parameter.name} to {dep.call} has no dependency marker,"
                        " no type annotation and no default value."
                        " This will produce a TypeError when this function is called."
                        " You must either provide a dependency marker, a type annotation or a default value."
                        f"\nPath: {get_path_str([*path, dep])}"
                    ),
                    path=[*path, dep],
                )
    return params


def check_task_scope_validity(
    task: Task,
    subtasks: Iterable[Task],
    scopes: Mapping[Scope, int],
    path: Iterable[DependantBase[Any]],
) -> None:
    if task.scope not in scopes:
        raise UnknownScopeError(
            f"Dependency{task.dependant} has an unknown scope {task.scope}."
            f"\nExample Path: {get_path_str(path)}"
        )
    for subtask in subtasks:
        if scopes[task.scope] < scopes[subtask.scope]:
            raise ScopeViolationError(
                f"{task.dependant.call} cannot depend on {subtask.dependant.call}"
                f" because {subtask.dependant.call}'s scope ({subtask.scope})"
                f" is narrower than {task.dependant.call}'s scope ({task.scope})"
                f"\nExample Path: {get_path_str(path)}"
            )


def build_task(
    dependency: DependantBase[Any],
    binds: Iterable[BindHook],
    tasks: Dict[CacheKey, Task],
    task_dag: Dict[Task, List[Task]],
    dependant_dag: Dict[DependantBase[Any], List[DependencyParameter]],
    path: Dict[DependantBase[Any], Any],
    scope_idxs: Mapping[Scope, int],
    scope_resolver: Optional[ScopeResolver],
) -> Task:

    call = dependency.call
    assert call is not None
    scope = dependency.scope

    if dependency.call in {d.call for d in path}:
        raise DependencyCycleError(
            "Dependencies are in a cycle",
            list(path.keys()),
        )

    params = get_params(dependency, binds, path)

    positional_parameters: "List[Task]" = []
    keyword_parameters: "Dict[str, Task]" = {}
    subtasks: "List[Task]" = []
    dep_params: "List[DependencyParameter]" = []

    path[dependency] = None  # any value will do, we only use the keys

    for param in params:
        dep_params.append(param)
        if param.dependency.call is not None:
            child_task = build_task(
                param.dependency,
                binds,
                tasks,
                task_dag,
                dependant_dag,
                path,
                scope_idxs,
                scope_resolver,
            )
            subtasks.append(child_task)
            if param.parameter is not None:
                if param.parameter.kind in POSITIONAL_PARAMS:
                    positional_parameters.append(child_task)
                else:
                    keyword_parameters[param.parameter.name] = child_task
        if (
            param.dependency not in dependant_dag
            and param.dependency.cache_key not in tasks
        ):
            dependant_dag[param.dependency] = []
    if scope_resolver:
        child_scopes = [st.scope for st in subtasks]
        scope = scope_resolver(dependency, child_scopes, tuple(scope_idxs.keys()))

    if dependency.cache_key in tasks:
        if tasks[dependency.cache_key].scope != scope:
            raise SolvingError(
                f"{dependency.call} was used with multiple scopes",
                path=list(path.keys()),
            )
        path.pop(dependency)
        return tasks[dependency.cache_key]

    task = Task(
        dependant=dependency,
        scope=scope,
        call=call,
        cache_key=dependency.cache_key,
        task_id=len(tasks),
        positional_parameters=positional_parameters,
        keyword_parameters=keyword_parameters,
        use_cache=dependency.use_cache,
    )
    dependant_dag[dependency] = dep_params
    tasks[dependency.cache_key] = task
    task_dag[task] = subtasks
    check_task_scope_validity(
        task,
        subtasks,
        scope_idxs,
        path,
    )
    # remove ourselves from the path
    path.pop(dependency)
    return task


def solve(
    dependency: DependantBase[T],
    scopes: Sequence[Scope],
    binds: Iterable[BindHook],
    scope_resolver: Optional[ScopeResolver],
) -> SolvedDependant[T]:
    """Solve a dependency.

    Returns a SolvedDependant that can be executed to get the dependency's value.
    """
    # If the dependency itself is a bind, replace it
    for hook in binds:
        match = hook(None, dependency)
        if match:
            dependency = match

    if dependency.call is None:  # pragma: no cover
        raise ValueError("DependantBase.call must not be None")

    task_dag: "Dict[Task, List[Task]]" = {}
    dep_dag: "Dict[DependantBase[Any], List[DependencyParameter]]" = {}
    scope_idxs = dict((scope, idx) for idx, scope in enumerate(scopes))

    # this is implemented recursively
    # which will crash on DAGs with depth > 1000 (default recursion limit)
    # if we encounter that in a real world use case
    # we can just rewrite this to be iterative
    root_task = build_task(
        dependency=dependency,
        binds=binds,
        tasks={},
        task_dag=task_dag,
        dependant_dag=dep_dag,
        # we use a dict to represent the path so that we can have
        # both O(1) lookups, and an ordered mutable sequence (via dict keys)
        # we simply ignore / don't use the dict values
        path={},
        scope_idxs=scope_idxs,
        scope_resolver=scope_resolver,
    )

    ts = TopologicalSorter(task_dag)
    static_order = tuple(ts.copy().static_order())
    ts.prepare()
    assert dependency.call is not None
    container_cache = SolvedDependantCache(
        root_task=root_task,
        topological_sorter=ts,
        static_order=static_order,
        empty_results=[None] * len(task_dag),
    )
    solved = SolvedDependant(
        dependency=dependency,
        dag=dep_dag,
        container_cache=container_cache,
    )
    return solved



import inspect
from typing import Any, Optional

from di._utils.inspect import get_type
from di.api.dependencies import DependantBase
from di.container._bind_hook import BindHook


def bind_by_type(
    provider: DependantBase[Any],
    dependency: type,
    *,
    covariant: bool = False,
) -> BindHook:
    """Hook to substitute the matched dependency"""

    def hook(
        param: Optional[inspect.Parameter], dependant: DependantBase[Any]
    ) -> Optional[DependantBase[Any]]:
        if dependant.call is dependency:
            return provider
        if param is None:
            return None
        type_annotation_option = get_type(param)
        if type_annotation_option is None:
            return None
        type_annotation = type_annotation_option.value
        if type_annotation is dependency:
            return provider
        if covariant:
            if inspect.isclass(type_annotation) and inspect.isclass(dependency):
                if dependency in type_annotation.__mro__:
                    return provider
        return None

    return hook



from typing import Any, List

from di.api.dependencies import DependantBase


class DependencyInjectionException(Exception):
    """Base exception for this library"""

    pass


class WiringError(DependencyInjectionException):
    """Raised when wiring (introspection into types) failed"""

    def __init__(self, msg: str, path: List[DependantBase[Any]]) -> None:
        super().__init__(msg)
        self.path = path


class UnknownScopeError(DependencyInjectionException):
    """Raised when a dependency to be executed has an unknown Scope"""


class DuplicateScopeError(DependencyInjectionException):
    """Raised when enter_scope() is called with an existing scope"""


class DependencyCycleError(DependencyInjectionException):
    """Raised when a dependency cycle is detected"""

    def __init__(self, msg: str, path: List[DependantBase[Any]]) -> None:
        super().__init__(msg)
        self.path = path


class ScopeViolationError(DependencyInjectionException):
    """Raised when Scope layering is violated.
    Using pytests' Scopes as an example, if A has "session" Scope and B has "function" Scope,
    A cannot depend on B (in fact, pytest will also throw an error).
    """


class SolvingError(DependencyInjectionException):
    """Raised when there is an issue solving, for example if a dependency appears twice with different scopes"""

    def __init__(self, msg: str, path: List[DependantBase[Any]]) -> None:
        super().__init__(msg)
        self.path = path


class IncompatibleDependencyError(DependencyInjectionException):
    """Raised when an async context manager dependency is executed in a sync Scope"""



from __future__ import annotations

import inspect
from typing import Any, List, Optional, TypeVar, overload

from di._utils.inspect import get_parameters, get_type
from di.api.dependencies import (
    CacheKey,
    DependantBase,
    DependencyParameter,
    InjectableClassProvider,
)
from di.api.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProvider,
    DependencyProviderType,
    GeneratorProvider,
)
from di.api.scopes import Scope
from di.typing import get_markers_from_annotation

_VARIABLE_PARAMETER_KINDS = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD,
)

T = TypeVar("T")


class Marker:
    """A dependency marker holds information about a dependency.

    Used to tell `di` how to construct another class.

    For example:

    ```py
    def endpoint(conn: Annotated[DBConn, Marker(inject_db, scope="request")]):
        ...
    ```

    Building your own `Marker` can be critical to enable nice functionality.
    You could for example create a custom `Marker` "Header" than knows how to construct a `str`
    from the headers of a request. Resulting in:

    ```py
    def endpoint(content_type: FromHeader[str]):
        ...
    ```

    See more in [dependency-markers](https://www.adriangb.com/di/latest/wiring/#dependency-markers).
    """

    call: Optional[DependencyProvider]
    dependency: Optional[Any]
    scope: Scope
    use_cache: bool
    wire: bool
    sync_to_thread: bool

    def __init__(
        self,
        call: Optional[DependencyProviderType[Any]] = None,
        *,
        scope: Scope = None,
        use_cache: bool = True,
        wire: bool = True,
        sync_to_thread: bool = False,
    ) -> None:
        # by default we assume that call and dependency are the same thing
        # but we don't enforce this on subclasses so that they can assign
        # arbitrary meaning to dependency, e.g. a class that implements some API
        # to produce a callable
        self.call = self.dependency = call
        self.scope = scope
        self.use_cache = use_cache
        self.wire = wire
        self.sync_to_thread = sync_to_thread

    def register_parameter(self, param: inspect.Parameter) -> DependantBase[Any]:
        """Hook to register the parameter this Dependant corresponds to.

        This can be used to inferr self.call from a type annotation (autowiring),
        or to just register the type annotation.

        This method can return the same or a new instance of a Dependant to avoid modifying itself.
        """
        call = self.call
        if call is None and param.default is not param.empty:

            def inject_default_value() -> Any:
                return param.default

            call = inject_default_value
        if call is None:
            annotation_type_option = get_type(param)
            if annotation_type_option is not None and inspect.isclass(
                annotation_type_option.value
            ):
                if issubclass(annotation_type_option.value, InjectableClassProvider):
                    return annotation_type_option.value.__di_dependency__(param)
                else:
                    # a class type, a callable class instance or a function
                    call = annotation_type_option.value
        return Dependant[Any](
            call=call,
            scope=self.scope,
            use_cache=self.use_cache,
            wire=self.wire,
            sync_to_thread=self.sync_to_thread,
        )


class Dependant(DependantBase[T]):
    """Connect dependencies together.

    A `Dependant` can have sub-dependencies (also `Dependant`s).
    The first argument is a `Callable`, which is used to find the
    sub-dependencies.

    Arguments:
        call: used to find subdependencies
        wire: if True then `call` is introspected to find sub-dependencies.
        sync_to_thread: if True synchronous dependencies are run in a separate thread to avoid blocking the event loop
        scope: the Scope for this dependency (see https://www.adriangb.com/di/latest/scopes/)
        marker: the Marker from which this Defendant was constructed. This is included only for introspection purposes.
    """

    call: Optional[DependencyProviderType[T]]
    wire: bool
    sync_to_thread: bool
    scope: Scope
    marker: Optional[Marker]

    @overload
    def __init__(
        self,
        call: Optional[AsyncGeneratorProvider[T]] = ...,
        *,
        marker: Optional[Marker] = ...,
        scope: Scope = ...,
        use_cache: bool = ...,
        wire: bool = ...,
        sync_to_thread: bool = ...,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CoroutineProvider[T]] = ...,
        *,
        marker: Optional[Marker] = ...,
        scope: Scope = ...,
        use_cache: bool = ...,
        wire: bool = ...,
        sync_to_thread: bool = ...,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[GeneratorProvider[T]] = ...,
        *,
        marker: Optional[Marker] = ...,
        scope: Scope = ...,
        use_cache: bool = ...,
        wire: bool = ...,
        sync_to_thread: bool = ...,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CallableProvider[T]] = None,
        *,
        marker: Optional[Marker] = ...,
        scope: Scope = ...,
        use_cache: bool = ...,
        wire: bool = ...,
        sync_to_thread: bool = ...,
    ) -> None:
        ...

    def __init__(
        self,
        call: Optional[DependencyProviderType[T]] = None,
        *,
        marker: Optional[Marker] = None,
        scope: Scope = None,
        use_cache: bool = True,
        wire: bool = True,
        sync_to_thread: bool = False,
    ) -> None:
        self.call = call
        self.scope = scope
        self.use_cache = use_cache
        self.wire = wire
        self.sync_to_thread = sync_to_thread
        self.marker = marker

    @property
    def cache_key(self) -> CacheKey:
        if self.use_cache is False or self.call is None:
            return (self.__class__, id(self))
        return (self.__class__, self.call)

    def get_dependencies(self) -> List[DependencyParameter]:
        """Collect all of our sub-dependencies as parameters"""
        if self.wire is False or self.call is None:
            return []
        res: "List[DependencyParameter]" = []
        for param in get_parameters(self.call).values():
            sub_dependant: DependantBase[Any]
            if param.kind in _VARIABLE_PARAMETER_KINDS:
                continue
            else:
                maybe_sub_dependant_marker = next(
                    get_markers_from_annotation(param.annotation, Marker), None
                )
                if maybe_sub_dependant_marker is not None:
                    sub_dependant = maybe_sub_dependant_marker.register_parameter(param)
                else:
                    sub_dependant = self.get_default_marker().register_parameter(param)
            res.append(DependencyParameter(dependency=sub_dependant, parameter=param))
        return res

    def get_default_marker(self) -> Marker:
        return Marker(scope=self.scope, use_cache=self.use_cache)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(call={self.call}, use_cache={self.use_cache})"
        )



from contextlib import AsyncExitStack, ExitStack
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    cast,
)

from graphlib2 import TopologicalSorter

from di._utils.scope_map import ScopeMap
from di._utils.task import ExecutionState, Task
from di._utils.types import CacheKey
from di.api.executor import SupportsTaskGraph
from di.api.executor import Task as SupportsTask
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.api.solved import SolvedDependant


class TaskGraph:
    __slots__ = ("_uncopied_ts", "_copied_ts", "_static_order")
    _copied_ts: Optional[TopologicalSorter[Task]]

    def __init__(
        self,
        ts: TopologicalSorter[Task],
        static_order: Iterable[Task],
    ) -> None:
        self._uncopied_ts = ts
        self._copied_ts = None
        self._static_order = static_order

    def get_ready(self) -> Iterable[Task]:
        if self._copied_ts is None:
            self._copied_ts = self._uncopied_ts.copy()
        return self._copied_ts.get_ready()

    def done(self, task: SupportsTask[ExecutionState]) -> None:
        if self._copied_ts is None:
            self._copied_ts = self._uncopied_ts.copy()
        self._copied_ts.done(cast(Task, task))

    def is_active(self) -> bool:
        if self._copied_ts is None:
            self._copied_ts = self._uncopied_ts.copy()
        return self._copied_ts.is_active()

    def static_order(self) -> Iterable[Task]:
        return self._static_order


class SolvedDependantCache(NamedTuple):
    """Private data that the Container attaches to SolvedDependant"""

    root_task: Task
    topological_sorter: TopologicalSorter[Task]
    static_order: Iterable[Task]
    empty_results: List[Any]


EMPTY_VALUES: Dict[DependencyProvider, Any] = {}


def plan_execution(
    stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]],
    cache: ScopeMap[CacheKey, Any],
    solved: SolvedDependant[Any],
    values: Optional[Mapping[DependencyProvider, Any]] = None,
) -> Tuple[List[Any], SupportsTaskGraph[ExecutionState], ExecutionState, Task,]:
    solved_dependency_cache: SolvedDependantCache = solved.container_cache
    results = solved_dependency_cache.empty_results.copy()
    if values is None:
        values = EMPTY_VALUES
    execution_state = ExecutionState(
        values=values,
        stacks=stacks,
        results=results,
        cache=cache,
    )
    ts = TaskGraph(
        solved_dependency_cache.topological_sorter,
        solved_dependency_cache.static_order,
    )
    return (
        results,
        ts,
        execution_state,
        solved_dependency_cache.root_task,
    )
